*! v.1.0 Automated Individualized Student Assessment
*! Stas Kolenikov, skolenik at gmail
program define aisa
  version 10.1

  * syntax: aisa function prjname

  syntax [anything] [using], ///
      [loglevel(real 0) ///
      seed(passthru) replace VARiables(passthru) IDLength(passthru) LASTfirst johndoe display /// roster options
      external(passthru) exact VERBose /// assignment and check options
      noMISSing SCOreonly /// gradebook option
      ]

  global AISA_loglevel = `loglevel'

  gettoken fn other : anything
  local lfn = length("`fn'")

  cap use aisa, clear
  if _rc {
     if "`fn'"!="roster" {
       di "{err}No AISA files found, may be you are in a wrong directory, or have not defined any projects yet"
       exit 601
     }
  }
  else {
    global AISA_name : char _dta[name]
    global AISA_dir  : char _dta[dir]
    di "{txt}Project : {res}$AISA_name"
    if "`anything'" == "" {
       li
       exit
    }
  }

  * but if `anything' is not empty, need to do something

  * possible functions
  if "`fn'" == substr("assignment",1,max(4,`lfn')) {
    * create assignment
    Assign `other', `external' `verbose'
  }
  else if "`fn'" == substr("check",1,max(4,`lfn')) {
    * check completed assignment
    Check `other', `external' `exact' `verbose' `replace'
  }
  else if "`fn'" == "data" {
    * create the data
    Data `other', `external' `verbose' `replace'
  }
  else if "`fn'" == "keys" {
    * create the data
    Keys `other', `verbose' `replace'
  }
  else if "`fn'" == substr("compile",1,max(5,`lfn')) {
    * compile the grades from a completed assignment
    Compile `other', `replace'
  }
  else if "`fn'" == substr("update",1,max(3, `lfn')) {
    * update the grades
    Update `other'
  }
  else if "`fn'" == "irt" {
    * adjust the grades
    IRT `other'
  }
  else if "`fn'" == "do" {
    * run all the do-files in the directory
    DO `other', `verbose'
  }
  else if "`fn'" == substr("gradebook",1,max(4, `lfn')) {
    * display the grades
    Gradebook `other' , `lastfirst' `missing' `scoreonly'
  }
  else if "`fn'" == substr("compute",1,max(5, `lfn')) {
    * compute a composite grade
    Compute `other'
  }
  else if "`fn'" == substr("lookup", 1, max(4, `lfn')) {
    * look a student up
    Lookup `other'
  }
  else if "`fn'" == "retake" {
    * retake a test, use with caution!
    * check, replace implemented instead
    Retake `other'
  }
  else if "`fn'" == substr("roster",1,max(4,`lfn')) {
    * create roster
    Roster `other' `using', `seed' `replace' `variables' `lastfirst' `johndoe' `idlength' `display'
  }
  else if "`fn'" == "list" {
    * list of students
    List `other'
  }
  else if "`fn'" == substr("cleanup",1,max(4,`lfn')) {
    * create roster
    CleanUp
  }
  else {
    * what was that??
    di as err "`fn': invalid function of aisa"
    CleanIt
    exit 198
  }

  CleanIt

end

pro def Assign
   * create an assignment

   syntax name, [external(str) VERBose]

   local assgn `namelist'

   if "`verbose'" ~= "" local verbose noisily

   cap log close
   log using $AISA_name-`assgn', replace

   qui use CourseID using $AISA_name-roster, clear
   cap noi mkdir `assgn'
   local rc1 = _rc
   cap noi confirm file data-`assgn'.do
   local rc2 = _rc
   cap noi confirm file key-`assgn'.do
   local rc3 = _rc
   * must end with no errors, do not create any files
   set more off
   if (`rc1'+`rc2'+`rc3' == 0) {
       if $AISA_loglevel > 0 {
          di "{txt}Creating an assignment... "
       }
       copy data-`assgn'.do `assgn'/data-`assgn'.do
       copy key-`assgn'.do `assgn'/key-`assgn'.do
       qui cd `assgn'
       local nn = _N
       cap postclose AISA_post
       forvalues i=1/`nn' {
          local seed = CourseID[`i']
          preserve
          set seed `seed'
          cap `verbose' do data-`assgn'
          * it is the instructor's responsibility to do something meaningful!
          * return the data in memory
          if _rc {
             di as err "{err}data-`assgn' returned an error for CourseID = `seed'"
          }
          else {
             label data "Assignment `assgn' for student # `seed'"
             save data-`assgn'-`seed'
             if "`external'" != "" {
                 cap noi save `"`external'/data-`assgn'-`seed'"'
                 if _rc {
                    di "{err}directory `external' not found, or files cannot be overwritten"
                    exit
                 }
             }
             postfile AISA_post int Question double Key double Tolerance using key-`assgn'-`seed'
             cap `verbose' do key-`assgn'
             if _rc {
                di as err "{err}key-`assgn' returned an error for CourseID = `seed'"
                cap postclose AISA_post
             }
             else {
                postclose AISA_post
                use key-`assgn'-`seed', clear
                sort Question
                save, replace
             }
             restore
          }
       }

   }

   qui cd ..

end

pro def Keys
   * create keys for an assignment

   syntax name, [VERBose replace]

   local assgn `namelist'

   if "`verbose'" ~= "" local verbose noisily

   qui use CourseID using $AISA_name-roster, clear
   cap cd `assgn'
   local rc1 = _rc
   cap noi confirm file key-`assgn'.do
   local rc3 = _rc
   * must end with no errors, do not create any files
   set more off
   if (`rc1'+`rc3' == 0) {
       if $AISA_loglevel > 0 {
          di "{txt}Creating keys for an assignment {res}`assgn'"
       }
       local nn = _N
       cap postclose AISA_post
       forvalues i=1/`nn' {
          local seed = CourseID[`i']
          preserve
          set seed `seed'
          cap `verbose' use data-`assgn'-`seed'
          if ! _rc {
            postfile AISA_post int Question double Key double Tolerance using key-`assgn'-`seed', `replace'
            cap `verbose' do key-`assgn'
            local keyrc = _rc
            if `keyrc' {
               if `keyrc'==1 {
                  * stop it, for goodness sake
                  cap postclose AISA_post
                  cd ..
                  exit 1
               }
               else {
                  di as err "{err}key-`assgn' returned an error for CourseID = `seed'"
                  cap postclose AISA_post
               }
            }
            else {
               postclose AISA_post
               use key-`assgn'-`seed', clear
               sort Question
               save, replace
            }
          }
          restore
       }

   }

   qui cd ..

end

pro def Data
   * create data for an assignment

   syntax name, [external(str) VERBose replace]

   local assgn `namelist'

   if "`verbose'" ~= "" local verbose noisily

   cap log close
   qui log using $AISA_name-`assgn', replace

   qui use CourseID using $AISA_name-roster, clear
   cap cd `assgn'
   local rc1 = _rc
   cap noi confirm file data-`assgn'.do
   local rc2 = _rc
   * must end with no errors, do not create any files
   set more off
   if (`rc1'+`rc2' == 0) {
       if $AISA_loglevel > 0 {
          di "{txt}Creating data for assignment {res}`assgn'"
       }
       local nn = _N
       cap postclose AISA_post
       forvalues i=1/`nn' {
          local seed = CourseID[`i']
          preserve
          set seed `seed'
          cap `verbose' do data-`assgn'
          * it is the instructor's responsibility to do something meaningful!
          * return the data in memory
          if _rc {
             di as err "{err}data-`assgn' returned an error for CourseID = `seed'"
          }
          else {
             label data "Assignment `assgn' for student # `seed'"
             cap noi save data-`assgn'-`seed', `replace'
             if _rc {
                qui cd ..
                exit
             }
             if "`external'" != "" {
                 cap noi save `"`external'/data-`assgn'-`seed'"', `replace'
                 if _rc {
                    di "{err}directory `external' not found, or files cannot be overwritten"
                    exit
                 }
             }
          }
          restore
       }

   }

   qui cd ..
   cap log close

end


pro def Compile
   * compile results of an assignment

   syntax anything, [replace]

   * local version of the roster
   qui use CourseID using $AISA_name-roster, clear
   local nn=_N
   forvalues i=1/`nn' {
      local next = CourseID[`i']
      local who `who' `next'
   }

   local assgn `anything'

   cap cd `assgn'
   if _rc {
      di "{err}directory `assgn' not found, check assignment name and current directory"
      CleanIt
      exit 170
   }

   clear
   set obs 0
   gen CourseID = .
   gen Question = .
   gen Correct = .
   cap noi save results-`assgn'-long, `replace'
   if _rc {
      CleanIt
      exit 602
   }

   * append all the existing results
   tokenize `who'
   while "`1'" != "" {
      cap append using result-`assgn'-`1'
      if _rc {
         di "{err}Warning: assignment `assgn' not found for student `1'"
      }
      mac shift
   }
   cap drop Answer Missed
   save results-`assgn'-long, replace



   * record a line to aisa.dta
   cd $AISA_dir
   qui {
      use aisa
      local obsno = _N + 1
      set obs `obsno'
      replace assignment = "`assgn'" in l
      save, replace
   }




end

pro def DO
   * run student do-files
   syntax anything, [verbose]

   gettoken assgn what : anything
   local what = trim("`what'")

   cap cd `assgn'
   if _rc {
      di "{err}check the assignment name and current directory"
      CleanIt
      exit 617
   }

   * anything is interpreted as a prefix
   local dofiles : dir . files "`what'*.do"

   noi di as inp `"Files: `dofiles'"'

   * run those files, one by one
   tokenize `"`dofiles'"'
   while "`1'" ~= "" {
     cap postclose `assgn'
     cap log close
     if "`verbose'" != "" {
        set more off
        do `1', nostop
        set more on
        more
        * does not respect the current MORE settings
     }
     else {
        di as inp _n ". run `1', nostop"
        run `1', nostop
     }
     mac shift
   }

   qui cd ..

end

pro def Check
   * check an assignment
   syntax anything, [verbose replace]

   gettoken assgn who : anything

   di "{txt}Checking assignment {res}`assgn'"

   if "`who'" == "" {
      local who _all
   }

   if "`who'" == "_all" {
      * get the list of names from the roster
      local who
      qui use CourseID using $AISA_name-roster, clear
      local nn=_N
      forvalues i=1/`nn' {
         local next = CourseID[`i']
         local who `who' `next'
      }
   }

   cap log close
   log using $AISA_name-check-`assgn', append

   if $AISA_loglevel > 0 {
      di "{txt}The list of students to check the grades:" _n "{res}`who'"
   }

   cap noi cd `assgn'
   if _rc {
      di "{err}check the assignment name and current directory"
      CleanIt
      exit 617
   }

   tokenize `who'
   while "`1'" != "" {
      local thefile : dir . files "*answers-`assgn'-`1'.dta"
      if `"`thefile'"' == "" {
         * try the blackboard renaming scheme
         local thefile : dir . files "*answers_`assgn'_`1'.dta"
      }
      if `: word count `thefile' ' > 1 {
          di as err "Warning: multiple submissions from student # `1' found "
          local thefile : word 1 of `thefile'
      }
      cap use `thefile', clear
      if _rc {
          di as err "Error: submission from student # `1' not found"
      }
      else {
          cap confirm variable Question Answer
          if _rc {
             di as err "Error: variables Question and Answer not found in student # `1' submission"
          }
          else {
             sort Question
             qui merge Question using key-`assgn'-`1'
             cap assert _merge == 3
             if _rc {
               di "{err}Warning: answers from student # `1' are not matching up"
               qui drop if ! inlist(_merge,2,3)
             }
             cap bysort Question : assert _N == 1
             if _rc {
               di "{err}Warning: multiple answers received from student # `1'"
               qui bysort Question : replace Answer = . if _N > 1
               qui bysort Question : keep if _n == 1
               * need to preserve the number of questions!
             }
             gen byte Correct = ( abs(Answer-Key) <= Tolerance )
             drop Key* Toler* _merge
             gen long CourseID = `1'
             gen float RawScore = sum(Correct)
             qui replace RawScore = 100*RawScore[_N]/_N
             qui gen float AdjScore = RawScore
             * problems missed
             qui gen str10 Missed = ""
             local nq = _N
             qui forvalues k=1/`nq' {
                if !Correct[`k'] {
                  if Missed == "" replace Missed = "`k'"
                             else replace Missed = Missed + ", `k'"
                }
             }
             cap save result-`assgn'-`1', `replace'
             if _rc {
                di "{err}cannot write to result-`assgn'-`1'"
             }
             else {
               if "`verbose'" != "" list Question - AdjScore
                  else di "{txt}Student # {res}`1'{txt} data responses processed successfully."
             }
          }
      }
      mac shift
   }
   qui cd ..

end

pro def Update
   * create an assignment
   di "{txt}Updating the grades... "

   syntax anything

   if `: word count `anything' ' != 1 {
      di "{err}bad input to aisa update"
      exit
   }

   * anything must be the assignment name
   cap cd `anything'
   if _rc {
      di "{err}Results for `anything' not found"
      exit
   }

   * create the long list of grades
   clear
   gen long CourseID = .
   qui save `anything'-results-wide, replace
   local filelist : dir . files "result-`anything'*.dta"
   tokenize `"`filelist'"'
   while `"`1'"' ~= "" {
     append using `1'
     mac shift
   }
   keep if Question == 1
   keep CourseID *Score Missed
   save `anything'-results-wide, replace
   sort CourseID

   qui cd ..
   merge CourseID using $AISA_name-grades

   drop _merge
   * although there should be some value in tabulating the number of students
   * who got their answers submitted

   cap drop `anything'*

   foreach x of varlist RawScore AdjScore Missed {
      ren `x' `anything'_`x'
   }

   label var `anything'_RawScore "Raw score for `anything'"
   label var `anything'_AdjScore "Adjusted score for `anything'"
   label var `anything'_Missed   "Questions missed on `anything'"
   note : "Results for `anything' updated on $S_DATE at $S_TIME"
   sort CourseID
   save $AISA_name-grades, replace

   * generate the HTML output
   file open topost using `anything'-results.html, write text replace
   file write topost "<html><head><title>Results for `anything' assignment</title></head>" _n
   file write topost "<body><h2>Results for `anything' assignment</h2>" _n
   file write topost "<table width = 60% align=center border=1><tr><td>Course ID<td>Raw score<td>Adjusted score<td>Questions missed" _n

   * cycle over students
   local nn = _N
   forvalues k=1/`nn' {
       local id = CourseID[`k']
       local rs = string( `anything'_RawScore[`k'], "%6.2f")
       local as = string( `anything'_AdjScore[`k'], "%6.2f")
       if `anything'_Missed[`k'] == "" local miss &nbsp;
                                  else local miss = `anything'_Missed[`k']
       file write topost "<tr><td>`id'<td>`rs'<td>`as'<td>`miss'" _n
   }
   file write topost "</table></body></html>" _n
   file close topost

   di "{txt}The results were output to {browse `anything'-results.html}"

end

pro def IRT
   * create an assignment
   di "{txt}Adjusting the grades... "
end

pro def Gradebook
   * create an assignment
   syntax [anything] , [ lastfirst nomissing scoreonly ]
   di "{txt}Printing a gradebook... "

   * `anything' should be the assignment name
   * over can be CourseID or LastName
   local lover = length("`over'")
   if "`lastfirst'" != "" {
     local over LastName FirstName
   }
   else {
     local over CourseID
   }

   use ${AISA_dir}/${AISA_name}-grades, clear

   local assgn `anything'

   if "`assgn'" != "" {
      cap confirm numeric var `assgn'_RawScore
      if _rc {
         di "{err}unknown `assgn'"
         exit
      }
      if "`missing'" == "nomissing" local touse if !mi( `assgn'_RawScore )
      if "`scoreonly'" == "" local assgn `assgn'*
      else local assgn `assgn'_RawScore
   }
   else {
      local assgn *Score
   }

   if "`lastfirst'" != "" {
      sort LastName
      local tolist FirstName LastName
   }
   list `tolist' CourseID `assgn' `touse'

end

pro def Compute
   * create an assignment
   di "{txt}Adding a composite grade... "
end

pro def Lookup
   * create an assignment
   di "{txt}Finding a student... "
   syntax anything
   * anything can be CourseID or LastName
   if `: word count `anything' ' > 1 {
      di "{err}Invalid input to aisa lookup. Need a student name or CourseID"
      exit
   }

   cap confirm number `anything'
   if _rc {
     * assume anything is LastName
     use $AISA_name-grades if LastName == "`anything'"
   }
   else {
     * assume anything is ID#
     use $AISA_name-grades if CourseID == "`anything'"
   }
   if _N == 0 {
     di "{err}Student `anything' not found"
     exit
   }

   list CourseID FirstName LastName *Score

end

pro def List
   use $AISA_name-roster
   li CourseID FirstName LastName
end

pro def Roster
   * create an assignment
   if $AISA_loglevel >0 {
      di as inp `"`0'"'
   }
   syntax anything using , seed(real) ///
       [ replace variables(str) idlength(real 6) lastfirst(str) johndoe ]

   clear

   if $AISA_loglevel > 0 {
     di "{txt}Name = {res}`name'{txt}, namelist = {res}`namelist'{txt}, anything = {res}`anything'"
   }

   qui {
     set obs 0
     global AISA_name `anything'
     global AISA_dir = c(pwd)
     char _dta[name] $AISA_name
     char _dta[seed] `seed'
     char _dta[dir]  $AISA_dir
     if $AISA_loglevel > 0 {
        di "{txt}Project : {res}$AISA_name"
     }
     gen str40 assignment = ""
     save aisa, `replace'
   }

   clear

   * try Stata format first
   cap use `using'
   if _rc {
     * if does not work, try something comma or tab separated
     if $AISA_loglevel > 0 di `"{inp} infile `variables' `using' "'
     cap infile `variables' `using'
     if _rc {
       * if does not work, relegate back to user
       di as err "don't know how to open {inp}`using' "
       exit 603
     }
   }

   * process name
   if "`lastfirst'" != "" {
      if $AISA_loglevel > 0 {
         di "{txt}Parsing the name..."
      }
      * parse the name
      qui {
        gen byte comma = strpos(`lastfirst',",")
        gen str40 LastName =  substr(`lastfirst',1,comma-1)
        gen str40 FirstName = substr(`lastfirst',comma+1,.)
      }
   }

   * generate new IDs
   set seed `seed'
   local base = 10^(`idlength'-1)
   gen long CourseID = mod( 9*`base'*uniform() + `base'*uniform()/1000, 9*`base') + `base'

   * check that all are unique
   cap bysort CourseID : assert _N == 1
   if _rc {
      * not all unique

      di "{err}DO SOMETHING!!!"
   }

   if "`johndoe'" != "" {
     * create a fictitious observation
     local nn = _N + 1
     qui {
       set obs `nn'
       replace LastName = "Doe" in l
       replace FirstName = "John" in l
       replace CourseID = `base'-1 in l
     }
   }

   sort CourseID

   label data "$AISA_name roster"
   save ${AISA_name}-roster, `replace'

   keep CourseID LastName FirstName
   save ${AISA_name}-grades, `replace'

end

pro def CleanIt
  cap log close
  qui cd $AISA_dir
  global AISA_loglevel
  global AISA_name
  global AISA_dir
  clear
end

exit

History:
Jan 20 -- work started: Roster, Assign, Check


Files:
  aisa.dta in the current directory: the name of the project = char[_dta]name
                                     the seed for the random numbers in the roster = char[_dta]seed
                                     the main directory      = char[_dta]dir
                                     assignment name and # of questions

  `name'-roster.dta in the current directory: the list of students with real ID and other school information



  assgn/data-assgn-#####.dta -- the data for the assignment
  assgn/key-assgn-#####.dta  -- the instructor's answers to the assignment
                                   Question: int question #
                                   Key_exact: the exact answer from r() or e() values
                                   Key_display: what the students are supposed to see on the screen
                                   Sorted by: Question
  assgn/answers-assgn-#####.dta -- student provided answers
                                   Question: int question #
                                   Answer: numeric answer entered by the student

Globals:
AISA_loglevel    = loglevel, default is 0 for no output
AISA_name        = name of the project
AISA_dir         = base directory

Debugs to perform:
  aisa check
    - not all questions answered
    - some questions answered several times
    - more questions than in the original assignment
